/* DOMC Document Object Model library in C
 * Copyright (c) 2001 Michael B. Allen <mballen@erols.com>
 *
 * The MIT License
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

/* expatls.c - DOM_DocumentLS
 */

#include <limits.h>
#include <ctype.h>
#include <expat.h>
#include <langinfo.h>
#ifdef USE_ENCDEC
  #include <encdec.h>
#endif
#include <mba/stack.h>
#include <mba/mbs.h>
#include "domc.h"

#ifndef strnlen
size_t strnlen (const char *s, size_t maxlen);
#endif
#ifndef strncpy
char *strncpy(char *dest, const char *src, size_t n);
#endif

struct user_data {
	DOM_String *buf;
	size_t siz;
	struct stack *stk;
	int cdata;
};

size_t
utf8tods(const char *src, size_t sn, struct user_data *ud)
{
	size_t n;

#ifdef USE_ENCDEC
	char *s;

	s = (char *)src;
	if ((n = dec_mbsncpy(&s, sn, NULL, -1, -1, "UTF-8")) == (size_t)-1) {
		DOM_Exception = DOM_CHARACTER_ENC_ERR;
		PMNO(DOM_Exception);
		return -1;
	}
#else
	n = strnlen(src, sn) + 1;
#endif
	if (n > ud->siz) {
		ud->siz = n > (2 * ud->siz) ? n : (2 * ud->siz);
		if ((ud->buf = realloc(ud->buf, ud->siz)) == NULL) {
			DOM_Exception = errno;
			PMNO(DOM_Exception);
			return -1;
		}
	}
#ifdef USE_ENCDEC
	if ((n = dec_mbsncpy(&s, sn, ud->buf, ud->siz, -1, "UTF-8")) == (size_t)-1) {
		DOM_Exception = DOM_CHARACTER_ENC_ERR;
		PMNO(DOM_Exception);
		return -1;
	}
#else
	strncpy(ud->buf, src, n);
	ud->buf[n - 1] = '\0';
#endif
	return n;
}

int
fputds(const DOM_String *s, FILE *stream)
{
	return fputs(s, stream);
}

static void
start_fn(void *userData, const XML_Char *name, const XML_Char **atts)
{
	struct user_data *ud = userData;
	DOM_Node *parent, *child;
	DOM_String *a;
	int i;

	if (DOM_Exception) {
		return;
	}

	if (ud == NULL || name == NULL || atts == NULL) {
		DOM_Exception = NULL_POINTER_ERR;
		return;
	}

	parent = stack_peek(ud->stk);
	if (parent == NULL) {
		DOM_Exception = NULL_POINTER_ERR;
		return;
	}
	if (utf8tods(name, -1, ud) == (size_t)-1) {
		AMSG("name=%s", name);
		return;
	}
	if ((child = DOM_Document_createElement(parent->ownerDocument, ud->buf)) == NULL) {
		AMSG("");
		return;
	}
	for (i = 0; atts[i]; i += 2) {
		if (utf8tods(atts[i], -1, ud) == (size_t)-1) {
			AMSG("");
			return;
		}
		a = mbsdup(ud->buf);
		if (utf8tods(atts[i + 1], -1, ud) == (size_t)-1) {
			AMSG("");
			return;
		}
		DOM_Element_setAttribute(child, a, ud->buf);
		free(a);
		if (DOM_Exception) {
			return;
		}
	}
	if (DOM_Node_appendChild(parent, child) == NULL) {
		AMSG("");
		return;
	}
	if (stack_push(ud->stk, child) == 0) {
		DOM_Exception = errno;
		AMSG("");
		return;
	}
}
static void
end_fn(void *userData, const XML_Char *name)
{
	stack_pop(((struct user_data *)userData)->stk);
}
static void
cdata_start_fn(void *userData)
{
	((struct user_data *)userData)->cdata = 1;
}
static void
cdata_end_fn(void *userData)
{
	((struct user_data *)userData)->cdata = 0;
}
static void
chardata_fn(void *userData, const XML_Char *s, int len)
{
	struct user_data *ud = userData;
	DOM_Text *tex;
	DOM_Node *parent;

	if (DOM_Exception) {
		return;
	}

	if (ud == NULL || s == NULL || len == 0) {
		DOM_Exception = NULL_POINTER_ERR;
		PMNO(DOM_Exception);
		return;
	}

	parent = stack_peek(ud->stk);
	if (parent == NULL) {
		DOM_Exception = NULL_POINTER_ERR;
		PMNO(DOM_Exception);
		return;
	}

	if (utf8tods(s, len, ud) == (size_t)-1) {
		AMSG("");
		return;
	}
	if (ud->cdata) {
		if ((tex = DOM_Document_createCDATASection(parent->ownerDocument, ud->buf)) == NULL) {
			AMSG("");
			return;
		}
	} else {
		if ((tex = DOM_Document_createTextNode(parent->ownerDocument, ud->buf)) == NULL) {
			AMSG("");
			return;
		}
	}

	DOM_Node_appendChild(parent, tex);
	if (DOM_Exception) {
		DOM_Document_destroyNode(parent->ownerDocument, tex);
	}
}
static void
comment_fn(void *userData, const XML_Char *s)
{
	struct user_data *ud = userData;
	DOM_Comment *com;
	DOM_Node *parent;

	if (DOM_Exception) {
		return;
	}

	parent = stack_peek(ud->stk);
	if (parent == NULL) {
		DOM_Exception = DOM_XML_PARSER_ERR;
		PMNO(DOM_Exception);
		return;
	}
	if (utf8tods(s, -1, ud) == (size_t)-1) {
		AMSG("");
		return;
	}
	if ((com = DOM_Document_createComment(parent->ownerDocument, ud->buf))) {
		DOM_Node_appendChild(parent, com);
		if (DOM_Exception) {
			DOM_Document_destroyNode(parent->ownerDocument, com);
		}
	}
}
static void
processing_fn(void *userData, const XML_Char *target, const XML_Char *data)
{
	struct user_data *ud = userData;
	DOM_ProcessingInstruction *pi;
	DOM_Node *parent;
	DOM_String *targ;

	if (DOM_Exception) {
		return;
	}

	parent = stack_peek(ud->stk);
	if (parent == NULL) {
		DOM_Exception = DOM_XML_PARSER_ERR;
		PMNO(DOM_Exception);
		return;
	}
	if (utf8tods(target, -1, ud) == (size_t)-1) {
		AMSG("");
		return;
	}
	targ = mbsdup(ud->buf);
	if (utf8tods(data, -1, ud) == (size_t)-1) {
		AMSG("");
		return;
	}
	if ((pi = DOM_Document_createProcessingInstruction(parent->ownerDocument, targ, ud->buf))) {
		DOM_Node_appendChild(parent, pi);
		if (DOM_Exception) {
			DOM_Document_destroyNode(parent->ownerDocument, pi);
		}
	}
	free(targ);
}
int
DOM_DocumentLS_fread(DOM_DocumentLS *doc, FILE *stream)
{
	XML_Parser *p;
	struct user_data ud;
	void *buf;
	size_t n;
	int ret, done;

	if (doc == NULL || stream == NULL) {
		DOM_Exception = NULL_POINTER_ERR;
		PMNF(DOM_Exception, ": doc=%p,stream=%p", doc, stream);
		return 0;
	}

	p = XML_ParserCreate(NULL);
	if (p == NULL) {
		DOM_Exception = DOM_XML_PARSER_ERR;
		PMNO(DOM_Exception);
		return 0;
	}

	ud.buf = NULL;
	ud.siz = 0;
	ud.stk = stack_new(500);
	ud.cdata = 0;

	doc->ownerDocument = doc;

	if (ud.stk == NULL || stack_push(ud.stk, doc) == 0) {
		DOM_Exception = DOM_XML_PARSER_ERR;
		PMNF(DOM_Exception, ": stk=%p", ud.stk);
		XML_ParserFree(p);
		stack_del(ud.stk, NULL);
		return 0;
	}

	XML_SetElementHandler(p, start_fn, end_fn);
	XML_SetCharacterDataHandler(p, chardata_fn);
	XML_SetCdataSectionHandler(p, cdata_start_fn, cdata_end_fn);
	XML_SetCommentHandler(p, comment_fn);
	XML_SetProcessingInstructionHandler(p , processing_fn);
	XML_SetUserData(p, &ud);

	ret = 1;
	for ( ;; ) {
		if ((buf = XML_GetBuffer(p, BUFSIZ)) == NULL) {
			DOM_Exception = DOM_XML_PARSER_ERR;
			PMNF(DOM_Exception, ": buf=NULL");
			ret = 0;
			break;
		}
		if ((n = fread(buf, 1, BUFSIZ, stream)) == 0 && ferror(stream)) {
			DOM_Exception = ferror(stream);
			PMNO(DOM_Exception);
			ret = 0;
			break;
		}
		if (XML_ParseBuffer(p, n, (done = feof(stream))) == 0 || DOM_Exception) {
			if (DOM_Exception == 0) {
				DOM_Exception = DOM_XML_PARSER_ERR;
				PMNF(DOM_Exception, ": %s: line %u",
							XML_ErrorString(XML_GetErrorCode(p)),
							XML_GetCurrentLineNumber(p));
			} else {
				AMSG("");
			}
			ret = 0;
			break;
		}
		if (done) {
			break;
		}
	}

	if (doc->u.Document.documentElement) {
		DOM_Element_normalize(doc->u.Document.documentElement);
	}

	free(ud.buf);
	stack_del(ud.stk, NULL);
	XML_ParserFree(p);

	doc->ownerDocument = NULL;

	return ret;
}
int
DOM_DocumentLS_load(DOM_DocumentLS *doc, const char *uri)
{
	FILE *fd;
	int ret;

	if (doc == NULL || uri == NULL) {
		DOM_Exception = NULL_POINTER_ERR;
		PMNF(DOM_Exception, ": doc=%p,uri=%s", doc, uri);
		return 0;
	}

	fd = fopen(uri, "r");
	if (fd == NULL) {
		DOM_Exception = errno;
		PMNF(DOM_Exception, ": uri=%s", uri);
		return 0;
	}

	ret = DOM_DocumentLS_fread(doc, fd);

	fclose(fd);

	return ret;
}

int
DOM_DocumentLS_fwrite(const DOM_DocumentLS *node, FILE *stream)
{
	NodeEntry *e;
	DOM_Node *c;

    if (node == NULL || stream == NULL) {
		DOM_Exception = NULL_POINTER_ERR;
		PMNF(DOM_Exception, ": node=%p,stream=%p", node, stream);
		return 0;
    }

	if (DOM_Exception) {
		return 0;
	}

    switch (node->nodeType) {
        case DOM_ELEMENT_NODE:
			fputc('<', stream);
			fputds(node->nodeName, stream);
			for (e = node->attributes->first; e != NULL; e = e->next) {
				fputc(' ', stream);
				fputds(e->node->nodeName, stream);
				fputs("=\"", stream);
				fputds(e->node->nodeValue, stream);
				fputc('"', stream);
			}
			if (DOM_Node_hasChildNodes(node)) {
				fputc('>', stream);
				for (c = node->firstChild; c != NULL; c = c->nextSibling) {
					if (DOM_DocumentLS_fwrite(c, stream) == 0) {
						/* Don't put msgno macro here or might overrun buf */
						return 0;
					}
            	}
				fputs("</", stream);
				fputds(node->nodeName, stream);
				fputc('>', stream);
			} else {
				fputs("/>", stream);
			}
            break;
        case DOM_ATTRIBUTE_NODE:
            break;
        case DOM_TEXT_NODE:
			fputds(node->nodeValue, stream);
            break;
        case DOM_CDATA_SECTION_NODE:
            break;
        case DOM_ENTITY_REFERENCE_NODE:
            break;
        case DOM_ENTITY_NODE:
            break;
        case DOM_PROCESSING_INSTRUCTION_NODE:
            break;
        case DOM_COMMENT_NODE:
			fputs("<!--", stream);
			fputds(node->nodeValue, stream);
			fputs("-->", stream);
            break;
        case DOM_DOCUMENT_NODE:
			fputs("<?xml version=\"1.0\" encoding=\"", stream);
			fputs(nl_langinfo(CODESET), stream);
			fputs("\"?>\n\n", stream);
            if (DOM_DocumentLS_fwrite(node->u.Document.documentElement, stream) == 0) {
				AMSG("");
				return 0;
			}
			fputc('\n', stream);
            break;
        case DOM_DOCUMENT_TYPE_NODE:
            break;
        case DOM_DOCUMENT_FRAGMENT_NODE:
            break;
        case DOM_NOTATION_NODE:
            break;
    }

	return DOM_Exception ? 0 : 1;
}
int
DOM_DocumentLS_save(DOM_DocumentLS *doc, const char *uri, const DOM_Node *node)
{
	FILE *fd;

	if ((doc == NULL && node == NULL) || uri == NULL) {
		DOM_Exception = NULL_POINTER_ERR;
		PMNF(DOM_Exception, ": doc=%p,uri=%s,node=%p", doc, uri, node);
		return 0;
	}

	fd = fopen(uri, "w");
	if (fd && DOM_DocumentLS_fwrite(doc ? doc : node, fd)) {
		fclose(fd);
		return 1;
	}
	DOM_Exception = errno;
	PMNF(DOM_Exception, ": uri=%s", uri);

	return 0;
}

